#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Peak Workflow management service commands."""

from typing import Any, Dict, List, Optional

import typer
from peak.cli import args, helpers
from peak.cli.args import DRY_RUN, GENERATE_YAML, OUTPUT_TYPES, PAGING
from peak.constants import OutputTypes, OutputTypesNoTable, OutputTypesOnlyJson
from peak.helpers import parse_list_of_strings
from peak.output import Writer
from peak.resources.workflows import Workflow
from typing_extensions import Annotated

app = typer.Typer(
    help="Create Machine Learning Pipelines using Workflows.",
    short_help="Create and manage Workflows.",
)

_WORKFLOW_STATUS = typer.Option(
    None,
    help="List of status to filter workflows. Valid values are `Draft`, `Running`, `Available`, `Paused`.",
)

_LAST_EXECUTION_STATUS = typer.Option(
    None,
    help="List of execution status of the latest run of the workflows to filter them by. Valid values are `Success`, `Running`, `Stopped`, `Stopping`, `Failed`.",
)

_LAST_MODIFIED_BY = typer.Option(None, help="List of users who last modified the workflow, used to filter workflows.")

_WORKFLOW_ID = typer.Argument(..., help="ID of the workflow to be used in this operation.")

_WORKFLOW_ID_OPTION = typer.Option(..., help="ID of the workflow to be used in this operation.")

_EXECUTION_ID = typer.Option(..., help="ID of the workflow execution to be used in this operation.")

_LIST_NAME = typer.Option(None, help="Workflow name to search for.")

_NAME = typer.Option(None, help="Name of the workflow")

_STEP_NAME = typer.Option(..., help="Name of the workflow step to be used in this operation.")

_REPOSITORY = typer.Option(None, help="URL of the repository containing the required files.")

_BRANCH = typer.Option(None, help="The branch of the repository to use.")

_TOKEN = typer.Option(None, help="The token to use to access the repository.")

_COMMAND = typer.Option(None, help="The command to run when workflow step is executed.")

_IMAGE_ID = typer.Option(None, help="The ID of the image to use for the workflow step.")

_IMAGE_VERSION_ID = typer.Option(None, help="The ID of the image version to use for the workflow step.")

_INSTANCE_TYPE_ID = typer.Option(None, help="The ID of the instance type to use for the workflow step.")

_STORAGE = typer.Option(None, help="The storage to use for the workflow step.")

_STEP_TIMEOUT = typer.Option(None, help="The time after which the step timeouts.")

_CLEAR_IMAGE_CACHE = typer.Option(None, help="Whether to clear image cache on workflow execution.")

_STEP_NAMES = typer.Option(None, help="List of step names to be updated.")

_EXECUTION_STATUS = typer.Option(
    None,
    help="List of workflow execution statuses. Valid values are `Success`, `Running`, `Stopped`, `Stopping` and `Failed`.",
)

_COUNT = typer.Option(
    None,
    help="Number of workflow executions required in the provided time range or 90 days. If not provided, all executions are returned.",
)


@app.command(short_help="Create a new workflow.")
def create(
    ctx: typer.Context,
    file: str = args.TEMPLATE_PATH,
    params_file: str = args.TEMPLATE_PARAMS_FILE,
    params: List[str] = args.TEMPLATE_PARAMS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Create*** a new workflow.

    \b
    🧩 ***Input file schema(yaml):***<br/>
    ```yaml
      body (map):
        name (string): Name of the worklfow.
        tags (list(map) | required: false):
            - name (string): Name of the tag.
        triggers (list(map) | required: false):
            - cron (string | required: false): A valid cron expression.
              webhook (boolean | required: false): Should be true if webhook type trigger is to be used.
        watchers (list(map) | required: false):
            - events (map):
                success (boolean | required: false): Whether to call event on success.
                fail (boolean | required: false): Whether to call event on failure.
                runtimeExceeded (int | required: false): The runtime in minutes after which event is called.
              user (string): User to be notified.
            - events (map):
                success (boolean | required: false): Whether to call event on success.
                fail (boolean | required: false): Whether to call event on failure.
                runtimeExceeded (int | required: false): The runtime in minutes after which event is called.
              webhook (map):
                name (string): Name of the webhook.
                url (string): URL of the webhook.
                payload (string): Webhook payload.
            - events (map):
                success (boolean | required: false): Whether to call event on success.
                fail (boolean | required: false): Whether to call event on failure.
                runtimeExceeded (int | required: false): The runtime in minutes after which event is called.
              email (map):
                name (string): Name of the email watcher.
                recipients (map):
                    to (list(str)): List of email addresses to send the email to. Email can be sent only to the users who are added in the tenant.
        retryOptions (map | required: false): # Workflow level retry options which will be applied to all steps.
            duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
            exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
            exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
            numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
        steps(map):
            <stepName> (map): # Dictionary containing the step configuration. Here the key is name of the step.
                # Standard Step
                imageId (int): ID of the existing image.
                imageVersionId: (int | required: false): ID of the existing image version.
                type (string | required: false): Type of workflow step. This should be 'standard' for a Standard Step.
                command (string): Command to run when step is executed.
                clearImageCache (boolean | required: false): Whether to clear image cache on workflow execution.
                stepTimeout (int | required: false): Time after which the step timeouts.
                parents (list(str) | required: false): List containing names of steps on which this step is dependent on.
                repository (map | required: false):
                    branch (string): Branch of the repository containing the required files.
                    token (string | required: false): The token to be used to clone the repository.
                    url (string): URL of the repository.
                resources (map | required: false):
                    instanceTypeId (int): ID of the instance type to be used in the step.
                    storage (string): Storage in GB. For example, "10GB".
                parameters (map | required: false):
                    env (map | required: false): Key-Value pair where key is the name of the env.
                    inherit (map | required: false): Key-Value pair where key is the name of the env and value is the name of the output parameter.
                    secrets (list(str) | required: false): List of secret names to be passed.
                outputParameters (map | required: false):
                    <keyName> (string | required: false): Represents a key-value pair where the key is the parameter name, and the value is the file name where the output is stored.
                executionParameters (map | required: false):
                    conditional (list(map) | required: false): # List of conditions to be evaluated based on the output params of the parent steps.
                        - condition (string): The evaluation condition, either 'equals' or 'not-equals'.
                          paramName (string): Name of the output parameter to be evaluated.
                          stepName (string): Name of the step containing the output parameter.
                          value (string): Value to be compared with the output parameter.
                    parentStatus (list(map) | required: false): # List of conditions to be evaluated based on the status of the parent steps.
                        - condition (string): The evaluation condition, either 'all-of' or 'one-of'.
                          parents (list(str)): List of parent step names.
                          status (list(str)): List of parent step statuses, with valid values being 'success' and 'failure'.
                runConfiguration (map | required: false):
                    retryOptions (map | required: false): # Step level retry options which will override the workflow level retry options.
                        duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
                        exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
                        exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
                        numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
                    skipConfiguration (map | required: false):
                        skip (boolean | required: false): Indicates whether to skip the step. Default is false.
                        skipDAG (boolean | required: false): Indicates whether to skip dependent steps, including the current step. Default is false.
                # HTTP Step
                type (string | required: true): Type of workflow step. This should be 'http' for a HTTP Step.
                method (string | required: true): The HTTP method to be used for the API call. Should be one of - get, post, put, patch, and delete
                url (string | required: true): The URL to make the HTTP API call to.
                payload (string | required: false): Stringified JSON payload to be sent in case of non-get requests.
                headers (map | required: false):
                    absolute (map | required: false): Key-Value pair where key is the name of the header and value is the value of the header.
                    secrets (map | required: false): Key-Value pair where key is the name of the header and value is the name of the secret to get the value from.
                auth (map | required: false):
                    type (string | required: true): The type of authentication to use. Should be one of - no-auth, oauth, basic, api-key, and bearer-token.
                    clientId (string | required: false): The Client ID used to get the access token in case of OAuth. Value should be the name of a secret. Required when type is oauth.
                    clientSecret (string | required: false): The Client Secret used to get the access token in case of OAuth. Value should be the name of a secret. Required when type is oauth.
                    authUrl (string | required: false): The URL that is hit to get the access token in case of OAuth. Required when type is oauth.
                    username (string | required: false): The username used for basic authentication. Required when type is basic.
                    password (string | required: false): The password used for basic authentication. Value should be the name of a secret. Required when type is basic.
                    apiKey (string | required: false): The API Key used for authentication. Value should be the name of a secret. Required when type is api-key.
                    bearerToken (string | required: false): The bearer token used for authentication. Value should be the name of a secret. Required when type is bearer-token.
                parameters (map | required: false):
                    env (map | required: false): Key-Value pair where key is the name of the env.
                    inherit (map | required: false): Key-Value pair where key is the name of the env and value is the name of the output parameter.
                    secrets (list(str) | required: false): List of secret names to be passed.
                outputParameters (map | required: false):
                    <keyName> (string | required: false): Represents a key-value pair where the key is the parameter name. Note that in case of HTTP step, value should always be "response.txt", otherwise this parameter would be rejected.
                executionParameters (map | required: false):
                    conditional (list(map) | required: false): # List of conditions to be evaluated based on the output params of the parent steps.
                        - condition (string): The evaluation condition, either 'equals' or 'not-equals'.
                          paramName (string): Name of the output parameter to be evaluated.
                          stepName (string): Name of the step containing the output parameter.
                          value (string): Value to be compared with the output parameter.
                    parentStatus (list(map) | required: false): # List of conditions to be evaluated based on the status of the parent steps.
                        - condition (string): The evaluation condition, either 'all-of' or 'one-of'.
                          parents (list(str)): List of parent step names.
                          status (list(str)): List of parent step statuses, with valid values being 'success' and 'failure'.
                runConfiguration (map | required: false):
                    retryOptions (map | required: false): # Step level retry options which will override the workflow level retry options.
                        duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
                        exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
                        exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
                        numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
                    skipConfiguration (map | required: false):
                        skip (boolean | required: false): Indicates whether to skip the step. Default is false.
                        skipDAG (boolean | required: false): Indicates whether to skip dependent steps, including the current step. Default is false.
                # Export Step
                type (string | required: true): Type of workflow step. This should be 'export' for a export Step.
                schema (string | required: true): Schema from which the data is to be exported.
                table (string | required: true): Table from which the data is to be exported.
                sortBy (string | required: true): Column name by which the data is to be sorted.
                sortOrder (string | required: false): Order in which the data is to be sorted. Should be one of - asc, desc. Default is asc.
                compression (boolean | required: false): Whether to compress the files to '.gz' format while exporting. Default is false.
                executionParameters (map | required: false):
                    conditional (list(map) | required: false): # List of conditions to be evaluated based on the output params of the parent steps.
                        - condition (string): The evaluation condition, either 'equals' or 'not-equals'.
                          paramName (string): Name of the output parameter to be evaluated.
                          stepName (string): Name of the step containing the output parameter.
                          value (string): Value to be compared with the output parameter.
                    parentStatus (list(map) | required: false): # List of conditions to be evaluated based on the status of the parent steps.
                        - condition (string): The evaluation condition, either 'all-of' or 'one-of'.
                          parents (list(str)): List of parent step names.
                          status (list(str)): List of parent step statuses, with valid values being 'success' and 'failure'.
                runConfiguration (map | required: false):
                    retryOptions (map | required: false): # Step level retry options which will override the workflow level retry options.
                        duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
                        exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
                        exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
                        numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
                    skipConfiguration (map | required: false):
                        skip (boolean | required: false): Indicates whether to skip the step. Default is false.
                        skipDAG (boolean | required: false): Indicates whether to skip dependent steps, including the current step. Default is false.
                # SQL Step
                type (string | required: true): Type of workflow step. This should be 'sql' for a sql Step.
                sqlQueryPath (string | required: false): Path to the SQL query file. One of the 'sqlQueryPath' or 'repository' is required.
                repository (map | required: false):
                    branch (string | required: true): Branch of the repository containing the required files.
                    token (string | required: false): The token to be used to clone the repository.
                    url (string | required: true): URL of the repository.
                    filePath (string | required: true): Path to the file containing the SQL query.
                parameters (map | required: false):
                    env (map | required: false): Key-Value pair where key is the name of the env.
                    inherit (map | required: false): Key-Value pair where key is the name of the env and value is the name of the output parameter.
                executionParameters (map | required: false):
                    conditional (list(map) | required: false): # List of conditions to be evaluated based on the output params of the parent steps.
                        - condition (string): The evaluation condition, either 'equals' or 'not-equals'.
                          paramName (string): Name of the output parameter to be evaluated.
                          stepName (string): Name of the step containing the output parameter.
                          value (string): Value to be compared with the output parameter.
                    parentStatus (list(map) | required: false): # List of conditions to be evaluated based on the status of the parent steps.
                        - condition (string): The evaluation condition, either 'all-of' or 'one-of'.
                          parents (list(str)): List of parent step names.
                          status (list(str)): List of parent step statuses, with valid values being 'success' and 'failure'.
                runConfiguration (map | required: false):
                    retryOptions (map | required: false): # Step level retry options which will override the workflow level retry options.
                        duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
                        exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
                        exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
                        numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
                    skipConfiguration (map | required: false):
                        skip (boolean | required: false): Indicates whether to skip the step. Default is false.
                        skipDAG (boolean | required: false): Indicates whether to skip dependent steps, including the current step. Default is false.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak workflows create /path/to/file.yml -v /path/to/params.yml
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "id": 123,
        "triggers": [
            {
                "webhook": "db88c21d-1add-45dd-a72e-8c6b83b68dee"
            }
        ]
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/create-workflow)
    """
    workflow_client: Workflow = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    body: Dict[str, Any] = helpers.template_handler(file=file, params_file=params_file, params=params)
    body = helpers.remove_unknown_args(body, workflow_client.create_workflow)
    with writer.pager():
        response: Dict[str, int] = workflow_client.create_workflow(**body)
        writer.write(response)


@app.command(short_help="Update an existing workflow.")
def update(
    ctx: typer.Context,
    workflow_id: int = _WORKFLOW_ID,
    file: str = args.TEMPLATE_PATH,
    params_file: str = args.TEMPLATE_PARAMS_FILE,
    params: List[str] = args.TEMPLATE_PARAMS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Update*** an existing workflow.

    \b
    🧩 ***Input file schema(yaml):***<br/>
    ```yaml
      body (map):
        name (string): Name of the worklfow.
        tags (list(map) | required: false):
            - name (string): Name of the tag.
        triggers (list(map) | required: false):
            - cron (string | required: false): A valid cron expression.
              webhook (boolean | required: false): Should be true if webhook type trigger is to be used.
        watchers (list(map) | required: false):
            - events (map):
                success (boolean | required: false): Whether to call event on success.
                fail (boolean | required: false): Whether to call event on failure.
                runtimeExceeded (int | required: false): The runtime in minutes after which event is called.
              user (string): User to be notified.
            - events (map):
                success (boolean | required: false): Whether to call event on success.
                fail (boolean | required: false): Whether to call event on failure.
                runtimeExceeded (int | required: false): The runtime in minutes after which event is called.
              webhook (map):
                name (string): Name of the webhook.
                url (string): URL of the webhook.
                payload (string): Webhook payload.
            - events (map):
                success (boolean | required: false): Whether to call event on success.
                fail (boolean | required: false): Whether to call event on failure.
                runtimeExceeded (int | required: false): The runtime in minutes after which event is called.
              email (map):
                name (string): Name of the email watcher.
                recipients (map):
                    to (list(str)): List of email addresses to send the email to. Email can be sent only to the users who are added in the tenant.
        retryOptions (map | required: false): # Workflow level retry options which will be applied to all steps.
            duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
            exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
            exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
            numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
        steps(map):
            <stepName> (map): # Dictionary containing the step configuration. Here the key is name of the step.
                # Standard Step
                imageId (int): ID of the existing image.
                imageVersionId: (int | required: false): ID of the existing image version.
                type (string | required: false): Type of workflow step. This should be 'standard' for Standard Step.
                command (string): Command to run when step is executed.
                clearImageCache (boolean | required: false): Whether to clear image cache on workflow execution.
                stepTimeout (int | required: false): Time after which the step timeouts.
                parents (list(str) | required: false): List containing names of steps on which this step is dependent on.
                repository (map | required: false):
                    branch (string): Branch of the repository containing the required files.
                    token (string | required: false): The token to be used to clone the repository.
                    url (string): URL of the repository.
                resources (map | required: false):
                    instanceTypeId (int): ID of the instance type to be used in the step.
                    storage (string): Storage in GB. For example, "10GB".
                parameters (map | required: false):
                    env (map | required: false): Key-Value pair where key is the name of the env.
                    inherit (map | required: false): Key-Value pair where key is the name of the env and value is the name of the output parameter.
                    secrets (list(str) | required: false): List of secret names to be passed.
                outputParameters (map | required: false):
                    <keyName> (string | required: false): Represents a key-value pair where the key is the parameter name, and the value is the file name where the output is stored.
                executionParameters (map | required: false):
                    conditional (list(map) | required: false): # List of conditions to be evaluated based on the output params of the parent steps.
                        - condition (string): The evaluation condition, either 'equals' or 'not-equals'.
                          paramName (string): Name of the output parameter to be evaluated.
                          stepName (string): Name of the step containing the output parameter.
                          value (string): Value to be compared with the output parameter.
                    parentStatus (list(map) | required: false): # List of conditions to be evaluated based on the status of the parent steps.
                        - condition (string): The evaluation condition, either 'all-of' or 'one-of'.
                          parents (list(str)): List of parent step names.
                          status (list(str)): List of parent step statuses, with valid values being 'success' and 'failure'.
                runConfiguration (map | required: false):
                    retryOptions (map | required: false): # Step level retry options which will override the workflow level retry options.
                        duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
                        exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
                        exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
                        numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
                    skipConfiguration (map | required: false):
                        skip (boolean | required: false): Indicates whether to skip the step. Default is false.
                        skipDAG (boolean | required: false): Indicates whether to skip dependent steps, including the current step. Default is false.
                # HTTP Step
                type (string | required: false): Type of workflow step. This should be 'http' for a HTTP Step.
                method (string | required: true): The HTTP method to be used for the API call. Should be one of - get, post, put, patch, and delete
                url (string | required: true): The URL to make the HTTP API call to.
                payload (string | required: false): Stringified JSON payload to be sent in case of non-get requests.
                headers (map | required: false):
                    absolute (map | required: false): Key-Value pair where key is the name of the header and value is the value of the header.
                    secrets (map | required: false): Key-Value pair where key is the name of the header and value is the name of the secret to get the value from.
                auth (map | required: false):
                    type (string | required: true): The type of authentication to use. Should be one of - no-auth, oauth, basic, api-key, and bearer-token.
                    clientId (string | required: false): The Client ID used to get the access token in case of OAuth. Value should be the name of a secret. Required when type is oauth.
                    clientSecret (string | required: false): The Client Secret used to get the access token in case of OAuth. Value should be the name of a secret. Required when type is oauth.
                    authUrl (string | required: false): The URL that is hit to get the access token in case of OAuth. Required when type is oauth.
                    username (string | required: false): The username used for basic authentication. Required when type is basic.
                    password (string | required: false): The password used for basic authentication. Value should be the name of a secret. Required when type is basic.
                    apiKey (string | required: false): The API Key used for authentication. Value should be the name of a secret. Required when type is api-key.
                    bearerToken (string | required: false): The bearer token used for authentication. Value should be the name of a secret. Required when type is bearer-token.
                parameters (map | required: false):
                    env (map | required: false): Key-Value pair where key is the name of the env.
                    inherit (map | required: false): Key-Value pair where key is the name of the env and value is the name of the output parameter.
                    secrets (list(str) | required: false): List of secret names to be passed.
                outputParameters (map | required: false):
                    <keyName> (string | required: false): Represents a key-value pair where the key is the parameter name. Note that in case of HTTP step, value should always be "response.txt", otherwise this parameter would be rejected.
                executionParameters (map | required: false):
                    conditional (list(map) | required: false): # List of conditions to be evaluated based on the output params of the parent steps.
                        - condition (string): The evaluation condition, either 'equals' or 'not-equals'.
                          paramName (string): Name of the output parameter to be evaluated.
                          stepName (string): Name of the step containing the output parameter.
                          value (string): Value to be compared with the output parameter.
                    parentStatus (list(map) | required: false): # List of conditions to be evaluated based on the status of the parent steps.
                        - condition (string): The evaluation condition, either 'all-of' or 'one-of'.
                          parents (list(str)): List of parent step names.
                          status (list(str)): List of parent step statuses, with valid values being 'success' and 'failure'.
                runConfiguration (map | required: false):
                    retryOptions (map | required: false): # Step level retry options which will override the workflow level retry options.
                        duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
                        exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
                        exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
                        numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
                    skipConfiguration (map | required: false):
                        skip (boolean | required: false): Indicates whether to skip the step. Default is false.
                        skipDAG (boolean | required: false): Indicates whether to skip dependent steps, including the current step. Default is false.
                # Export Step
                type (string | required: true): Type of workflow step. This should be 'export' for a export Step.
                schema (string | required: true): Schema from which the data is to be exported.
                table (string | required: true): Table from which the data is to be exported.
                sortBy (string | required: true): Column name by which the data is to be sorted.
                sortOrder (string | required: false): Order in which the data is to be sorted. Should be one of - asc, desc. Default is asc.
                compression (boolean | required: false): Whether to compress the files to '.gz' format while exporting. Default is false.
                executionParameters (map | required: false):
                    conditional (list(map) | required: false): # List of conditions to be evaluated based on the output params of the parent steps.
                        - condition (string): The evaluation condition, either 'equals' or 'not-equals'.
                          paramName (string): Name of the output parameter to be evaluated.
                          stepName (string): Name of the step containing the output parameter.
                          value (string): Value to be compared with the output parameter.
                    parentStatus (list(map) | required: false): # List of conditions to be evaluated based on the status of the parent steps.
                        - condition (string): The evaluation condition, either 'all-of' or 'one-of'.
                          parents (list(str)): List of parent step names.
                          status (list(str)): List of parent step statuses, with valid values being 'success' and 'failure'.
                runConfiguration (map | required: false):
                    retryOptions (map | required: false): # Step level retry options which will override the workflow level retry options.
                        duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
                        exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
                        exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
                        numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
                    skipConfiguration (map | required: false):
                        skip (boolean | required: false): Indicates whether to skip the step. Default is false.
                        skipDAG (boolean | required: false): Indicates whether to skip dependent steps, including the current step. Default is false.
                # SQL Step
                type (string | required: true): Type of workflow step. This should be 'sql' for a sql Step.
                sqlQueryPath (string | required: false): Path to the SQL query file. One of the 'sqlQueryPath' or 'repository' is required.
                repository (map | required: false):
                    branch (string | required: true): Branch of the repository containing the required files.
                    token (string | required: false): The token to be used to clone the repository.
                    url (string | required: true): URL of the repository.
                    filePath (string | required: true): Path to the file containing the SQL query.
                parameters (map | required: false):
                    env (map | required: false): Key-Value pair where key is the name of the env.
                    inherit (map | required: false): Key-Value pair where key is the name of the env and value is the name of the output parameter.
                executionParameters (map | required: false):
                    conditional (list(map) | required: false): # List of conditions to be evaluated based on the output params of the parent steps.
                        - condition (string): The evaluation condition, either 'equals' or 'not-equals'.
                          paramName (string): Name of the output parameter to be evaluated.
                          stepName (string): Name of the step containing the output parameter.
                          value (string): Value to be compared with the output parameter.
                    parentStatus (list(map) | required: false): # List of conditions to be evaluated based on the status of the parent steps.
                        - condition (string): The evaluation condition, either 'all-of' or 'one-of'.
                          parents (list(str)): List of parent step names.
                          status (list(str)): List of parent step statuses, with valid values being 'success' and 'failure'.
                runConfiguration (map | required: false):
                    retryOptions (map | required: false): # Step level retry options which will override the workflow level retry options.
                        duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
                        exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
                        exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
                        numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
                    skipConfiguration (map | required: false):
                        skip (boolean | required: false): Indicates whether to skip the step. Default is false.
                        skipDAG (boolean | required: false): Indicates whether to skip dependent steps, including the current step. Default is false.
      ```

    \b
    📝 ***Example usage:***
    ```bash
    peak workflows update 9999 /path/to/file.yml -v /path/to/params.yml
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "id": 123,
        "triggers": [
            {
                "webhook": "db88c21d-1add-45dd-a72e-8c6b83b68dee"
            }
        ]
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/update-workflow)
    """
    workflow_client: Workflow = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    body: Dict[str, Any] = helpers.template_handler(file=file, params_file=params_file, params=params)
    body = helpers.remove_unknown_args(body, workflow_client.update_workflow)
    with writer.pager():
        response: Dict[str, Any] = workflow_client.update_workflow(workflow_id=workflow_id, **body)
        writer.write(response)


@app.command(
    short_help="Create a new workflow or Update an existing workflow.",
)
def create_or_update(
    ctx: typer.Context,
    file: str = args.TEMPLATE_PATH,
    params_file: str = args.TEMPLATE_PARAMS_FILE,
    params: List[str] = args.TEMPLATE_PARAMS,
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Create*** a new workflow or ***Update*** an existing workflow based on workflow name.

    \b
    🧩 ***Input file schema(yaml):***<br/>
    ```yaml
      body (map):
        name (string): Name of the worklfow.
        tags (list(map) | required: false):
            - name (string): Name of the tag.
        triggers (list(map) | required: false):
            - cron (string | required: false): A valid cron expression.
              webhook (boolean | required: false): Should be true if webhook type trigger is to be used.
        watchers (list(map) | required: false):
            - events (map):
                success (boolean | required: false): Whether to call event on success.
                fail (boolean | required: false): Whether to call event on failure.
                runtimeExceeded (int | required: false): The runtime in minutes after which event is called.
              user (string): User to be notified.
            - events (map):
                success (boolean | required: false): Whether to call event on success.
                fail (boolean | required: false): Whether to call event on failure.
                runtimeExceeded (int | required: false): The runtime in minutes after which event is called.
              webhook (map):
                name (string): Name of the webhook.
                url (string): URL of the webhook.
                payload (string): Webhook payload.
            - events (map):
                success (boolean | required: false): Whether to call event on success.
                fail (boolean | required: false): Whether to call event on failure.
                runtimeExceeded (int | required: false): The runtime in minutes after which event is called.
              email (map):
                name (string): Name of the email watcher.
                recipients (map):
                    to (list(str)): List of email addresses to send the email to. Email can be sent only to the users who are added in the tenant.
        retryOptions (map | required: false): # Workflow level retry options which will be applied to all steps.
            duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
            exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
            exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
            numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
        steps(map):
            <stepName> (map): # Dictionary containing the step configuration. Here the key is name of the step.
                # Standard Step
                imageId (int): ID of the existing image.
                imageVersionId: (int | required: false): ID of the existing image version.
                type (string | required: false): Type of workflow step. This should be 'standard' for Standard Step.
                command (string): Command to run when step is executed.
                clearImageCache (boolean | required: false): Whether to clear image cache on workflow execution.
                stepTimeout (int | required: false): Time after which the step timeouts.
                parents (list(str) | required: false): List containing names of steps on which this step is dependent on.
                repository (map | required: false):
                    branch (string): Branch of the repository containing the required files.
                    token (string | required: false): The token to be used to clone the repository.
                    url (string): URL of the repository.
                resources (map | required: false):
                    instanceTypeId (int): ID of the instance type to be used in the step.
                    storage (string): Storage in GB. For example, "10GB".
                parameters (map | required: false):
                    env (map | required: false): Key-Value pair where key is the name of the env.
                    inherit (map | required: false): Key-Value pair where key is the name of the env and value is the name of the output parameter.
                    secrets (list(str) | required: false): List of secret names to be passed.
                outputParameters (map | required: false):
                    <keyName> (string | required: false): Represents a key-value pair where the key is the parameter name, and the value is the file name where the output is stored.
                executionParameters (map | required: false):
                    conditional (list(map) | required: false): # List of conditions to be evaluated based on the output params of the parent steps.
                        - condition (string): The evaluation condition, either 'equals' or 'not-equals'.
                          paramName (string): Name of the output parameter to be evaluated.
                          stepName (string): Name of the step containing the output parameter.
                          value (string): Value to be compared with the output parameter.
                    parentStatus (list(map) | required: false): # List of conditions to be evaluated based on the status of the parent steps.
                        - condition (string): The evaluation condition, either 'all-of' or 'one-of'.
                          parents (list(str)): List of parent step names.
                          status (list(str)): List of parent step statuses, with valid values being 'success' and 'failure'.
                runConfiguration (map | required: false):
                    retryOptions (map | required: false): # Step level retry options which will override the workflow level retry options.
                        duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
                        exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
                        exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
                        numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
                    skipConfiguration (map | required: false):
                        skip (boolean | required: false): Indicates whether to skip the step. Default is false.
                        skipDAG (boolean | required: false): Indicates whether to skip dependent steps, including the current step. Default is false.
                # HTTP Step
                type (string | required: false): Type of workflow step. This should be 'http' for a HTTP Step.
                method (string | required: true): The HTTP method to be used for the API call. Should be one of - get, post, put, patch, and delete
                url (string | required: true): The URL to make the HTTP API call to.
                payload (string | required: false): Stringified JSON payload to be sent in case of non-get requests.
                headers (map | required: false):
                    absolute (map | required: false): Key-Value pair where key is the name of the header and value is the value of the header.
                    secrets (map | required: false): Key-Value pair where key is the name of the header and value is the name of the secret to get the value from.
                auth (map | required: false):
                    type (string | required: true): The type of authentication to use. Should be one of - no-auth, oauth, basic, api-key, and bearer-token.
                    clientId (string | required: false): The Client ID used to get the access token in case of OAuth. Value should be the name of a secret. Required when type is oauth.
                    clientSecret (string | required: false): The Client Secret used to get the access token in case of OAuth. Value should be the name of a secret. Required when type is oauth.
                    authUrl (string | required: false): The URL that is hit to get the access token in case of OAuth. Required when type is oauth.
                    username (string | required: false): The username used for basic authentication. Required when type is basic.
                    password (string | required: false): The password used for basic authentication. Value should be the name of a secret. Required when type is basic.
                    apiKey (string | required: false): The API Key used for authentication. Value should be the name of a secret. Required when type is api-key.
                    bearerToken (string | required: false): The bearer token used for authentication. Value should be the name of a secret. Required when type is bearer-token.
                parameters (map | required: false):
                    env (map | required: false): Key-Value pair where key is the name of the env.
                    inherit (map | required: false): Key-Value pair where key is the name of the env and value is the name of the output parameter.
                    secrets (list(str) | required: false): List of secret names to be passed.
                outputParameters (map | required: false):
                    <keyName> (string | required: false): Represents a key-value pair where the key is the parameter name. Note that in case of HTTP step, value should always be "response.txt", otherwise this parameter would be rejected.
                executionParameters (map | required: false):
                    conditional (list(map) | required: false): # List of conditions to be evaluated based on the output params of the parent steps.
                        - condition (string): The evaluation condition, either 'equals' or 'not-equals'.
                          paramName (string): Name of the output parameter to be evaluated.
                          stepName (string): Name of the step containing the output parameter.
                          value (string): Value to be compared with the output parameter.
                    parentStatus (list(map) | required: false): # List of conditions to be evaluated based on the status of the parent steps.
                        - condition (string): The evaluation condition, either 'all-of' or 'one-of'.
                          parents (list(str)): List of parent step names.
                          status (list(str)): List of parent step statuses, with valid values being 'success' and 'failure'.
                runConfiguration (map | required: false):
                    retryOptions (map | required: false): # Step level retry options which will override the workflow level retry options.
                        duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
                        exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
                        exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
                        numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
                    skipConfiguration (map | required: false):
                        skip (boolean | required: false): Indicates whether to skip the step. Default is false.
                        skipDAG (boolean | required: false): Indicates whether to skip dependent steps, including the current step. Default is false.
                # Export Step
                type (string | required: true): Type of workflow step. This should be 'export' for a export Step.
                schema (string | required: true): Schema from which the data is to be exported.
                table (string | required: true): Table from which the data is to be exported.
                sortBy (string | required: true): Column name by which the data is to be sorted.
                sortOrder (string | required: false): Order in which the data is to be sorted. Should be one of - asc, desc. Default is asc.
                compression (boolean | required: false): Whether to compress the files to '.gz' format while exporting. Default is false.
                executionParameters (map | required: false):
                    conditional (list(map) | required: false): # List of conditions to be evaluated based on the output params of the parent steps.
                        - condition (string): The evaluation condition, either 'equals' or 'not-equals'.
                          paramName (string): Name of the output parameter to be evaluated.
                          stepName (string): Name of the step containing the output parameter.
                          value (string): Value to be compared with the output parameter.
                    parentStatus (list(map) | required: false): # List of conditions to be evaluated based on the status of the parent steps.
                        - condition (string): The evaluation condition, either 'all-of' or 'one-of'.
                          parents (list(str)): List of parent step names.
                          status (list(str)): List of parent step statuses, with valid values being 'success' and 'failure'.
                runConfiguration (map | required: false):
                    retryOptions (map | required: false): # Step level retry options which will override the workflow level retry options.
                        duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
                        exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
                        exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
                        numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
                    skipConfiguration (map | required: false):
                        skip (boolean | required: false): Indicates whether to skip the step. Default is false.
                        skipDAG (boolean | required: false): Indicates whether to skip dependent steps, including the current step. Default is false.
                # SQL Step
                type (string | required: true): Type of workflow step. This should be 'sql' for a sql Step.
                sqlQueryPath (string | required: false): Path to the SQL query file. One of the 'sqlQueryPath' or 'repository' is required.
                repository (map | required: false):
                    branch (string | required: true): Branch of the repository containing the required files.
                    token (string | required: false): The token to be used to clone the repository.
                    url (string | required: true): URL of the repository.
                    filePath (string | required: true): Path to the file containing the SQL query.
                parameters (map | required: false):
                    env (map | required: false): Key-Value pair where key is the name of the env.
                    inherit (map | required: false): Key-Value pair where key is the name of the env and value is the name of the output parameter.
                executionParameters (map | required: false):
                    conditional (list(map) | required: false): # List of conditions to be evaluated based on the output params of the parent steps.
                        - condition (string): The evaluation condition, either 'equals' or 'not-equals'.
                          paramName (string): Name of the output parameter to be evaluated.
                          stepName (string): Name of the step containing the output parameter.
                          value (string): Value to be compared with the output parameter.
                    parentStatus (list(map) | required: false): # List of conditions to be evaluated based on the status of the parent steps.
                        - condition (string): The evaluation condition, either 'all-of' or 'one-of'.
                          parents (list(str)): List of parent step names.
                          status (list(str)): List of parent step statuses, with valid values being 'success' and 'failure'.
                runConfiguration (map | required: false):
                    retryOptions (map | required: false): # Step level retry options which will override the workflow level retry options.
                        duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
                        exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
                        exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
                        numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
                    skipConfiguration (map | required: false):
                        skip (boolean | required: false): Indicates whether to skip the step. Default is false.
                        skipDAG (boolean | required: false): Indicates whether to skip dependent steps, including the current step. Default is false.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak workflows create-or-update /path/to/file.yml -v /path/to/params.yml
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "id": 123,
        "triggers": [
            {
                "webhook": "db88c21d-1add-45dd-a72e-8c6b83b68dee"
            }
        ]
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/create-workflow)
    """
    workflow_client: Workflow = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    body: Dict[str, Any] = helpers.template_handler(file=file, params_file=params_file, params=params)
    body = helpers.remove_unknown_args(body, workflow_client.create_or_update_workflow)
    with writer.pager():
        response: Dict[str, int] = workflow_client.create_or_update_workflow(**body)
        writer.write(response)


@app.command(short_help="Update required fields of an existing workflow.")
def patch(
    ctx: typer.Context,
    workflow_id: int = _WORKFLOW_ID,
    file: Annotated[
        Optional[str],
        typer.Argument(
            ...,
            help="Path to the file that defines the body for this operation, supports both `yaml` file or a `jinja` template.",
        ),
    ] = None,
    params_file: str = args.TEMPLATE_PARAMS_FILE,
    params: List[str] = args.TEMPLATE_PARAMS,
    name: str = _NAME,
    repository: str = _REPOSITORY,
    branch: str = _BRANCH,
    token: str = _TOKEN,
    image_id: int = _IMAGE_ID,
    image_version_id: int = _IMAGE_VERSION_ID,
    command: str = _COMMAND,
    clear_image_cache: bool = _CLEAR_IMAGE_CACHE,
    step_timeout: int = _STEP_TIMEOUT,
    instance_type_id: int = _INSTANCE_TYPE_ID,
    storage: str = _STORAGE,
    step_names: List[str] = _STEP_NAMES,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Update*** an existing workflow.

    \b
    This command allows to efficiently modify trigger details, watchers, workflow name, and specific step attributes such as repository URL, branch, token, image ID, version ID etc.

    \b
    By specifying step_names, we can globally update specified steps with provided parameters, streamlining the update process. If step_names is not provided, all the steps for that workflow would be updated.

    \b
    Alternatively, we can selectively modify individual step attributes across different steps by providing the details in the yaml file. With this, we can also add new steps to the workflow by providing the parameters required by the step.

    \b
    If both body and specific parameters are used, the latter takes precedence.

    \b
    🧩 ***Input file schema(yaml):***<br/>
    ```yaml
    body (map | required: false):
        name (string | required: false): Name of the worklfow.
        tags (list(map) | required: false):
            - name (string): Name of the tag.
        triggers (list(map) | required: false):
            - cron (string | required: false): A valid cron expression.
              webhook (boolean | required: false): Should be true if webhook type trigger is to be used.
        watchers (list(map) | required: false):
            - events (map):
                success (boolean | required: false): Whether to call event on success.
                fail (boolean | required: false): Whether to call event on failure.
                runtimeExceeded (int | required: false): The runtime in minutes after which event is called.
              user (string): User to be notified.
            - events (map):
                success (boolean | required: false): Whether to call event on success.
                fail (boolean | required: false): Whether to call event on failure.
                runtimeExceeded (int | required: false): The runtime in minutes after which event is called.
              webhook (map):
                name (string): Name of the webhook.
                url (string): URL of the webhook.
                payload (string): Webhook payload.
            - events (map):
                success (boolean | required: false): Whether to call event on success.
                fail (boolean | required: false): Whether to call event on failure.
                runtimeExceeded (int | required: false): The runtime in minutes after which event is called.
              email (map):
                name (string): Name of the email watcher.
                recipients (map):
                    to (list(str)): List of email addresses to send the email to. Email can be sent only to the users who are added in the tenant.
        retryOptions (map | required: false): # Workflow level retry options which will be applied to all steps.
            duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
            exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
            exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
            numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
        steps(map):
            <stepName> (map): # Dictionary containing the step configuration. Here the key is name of the step.

            # Standard Step
                imageId (int): ID of the existing image.
                imageVersionId: (int | required: false): ID of the existing image version.
                type (string | required: false): Type of workflow step. This should be standard for Standard Step.
                command (string): Command to run when step is executed.
                clearImageCache (boolean | required: false): Whether to clear image cache on workflow execution.
                stepTimeout (int | required: false): Time after which the step timeouts.
                parents (list(str) | required: false): List containing names of steps on which this step is dependent on.
                repository (map | required: false):
                    branch (string): Branch of the repository containing the required files.
                    token (string | required: false): The token to be used to clone the repository.
                    url (string): URL of the repository.
                resources (map | required: false):
                    instanceTypeId (int): ID of the instance type to be used in the step.
                    storage (string): Storage in GB. For example, "10GB".
                parameters (map | required: false):
                    env (map | required: false): Key-Value pair where key is the name of the env.
                    secrets (list(str) | required: false): List of secret names to be passed.
                outputParameters (map | required: false):
                    <keyName> (string | required: false): Represents a key-value pair where the key is the parameter name, and the value is the file name where the output is stored.
                executionParameters (map | required: false):
                    conditional (list(map) | required: false): # List of conditions to be evaluated based on the output params of the parent steps.
                        - condition (string): The evaluation condition, either 'equals' or 'not-equals'.
                          paramName (string): Name of the output parameter to be evaluated.
                          stepName (string): Name of the step containing the output parameter.
                          value (string): Value to be compared with the output parameter.
                    parentStatus (list(map) | required: false): # List of conditions to be evaluated based on the status of the parent steps.
                        - condition (string): The evaluation condition, either 'all-of' or 'one-of'.
                          parents (list(str)): List of parent step names.
                          status (list(str)): List of parent step statuses, with valid values being 'success' and 'failure'.
                runConfiguration (map | required: false):
                    retryOptions (map | required: false): # Step level retry options which will override the workflow level retry options.
                        duration (int | required: false): Duration in seconds after which the step is retried if it fails. Default is 5 seconds and maximum is 120 seconds.
                        exitCodes (list(int) | required: false): List of exit codes for which the step is retried. If not provided, the step is retried for all exit codes.
                        exponentialBackoff (boolean | required: false): Whether to use exponential backoff for retrying the step. If provided as true, the factor used will be 2.
                        numberOfRetries (int | required: false): Number of times the step should be retried. Default is 3 and maximum is 5.
                    skipConfiguration (map | required: false):
                        skip (boolean | required: false): Indicates whether to skip the step. Default is false.
                        skipDAG (boolean | required: false): Indicates whether to skip dependent steps, including the current step. Default is false.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak workflows patch 9999 /path/to/file.yml -v /path/to/params.yml
    ```

    \b
    📝 ***Example usage for updating workflow by passing only required parameters:***
    ```bash
    peak workflows patch 9999 --name <workflow_name> --image-id <image-id> --version-id <image-version-id> --step-names <step-name-1>,<step-name-2>
    ```

    \b
    🆗 ***Response:***
    ```js
    {
        "triggers": [
            {
                "webhook": "db88c21d-1add-45dd-a72e-8c6b83b68dee"
            }
        ]
        "id": 9999,
    }
    ```
    """
    workflow_client: Workflow = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]
    body: Dict[str, Any] = {}

    if file:
        body = helpers.template_handler(file=file, params_file=params_file, params=params)
        body = helpers.remove_unknown_args(body, workflow_client.patch_workflow)

    with writer.pager():
        response: Dict[str, Any] = workflow_client.patch_workflow(
            workflow_id=workflow_id,
            body=body["body"] if body else {},
            step_names=step_names,
            name=name,
            repository=repository,
            branch=branch,
            token=token,
            image_id=image_id,
            image_version_id=image_version_id,
            command=command,
            clear_image_cache=clear_image_cache,
            step_timeout=step_timeout,
            instance_type_id=instance_type_id,
            storage=storage,
        )
        writer.write(response)


@app.command("list", short_help="List workflows.")
def list_workflows(
    ctx: typer.Context,
    page_size: Optional[int] = args.PAGE_SIZE,
    page_number: Optional[int] = args.PAGE_NUMBER,
    workflow_status: Optional[List[str]] = _WORKFLOW_STATUS,
    last_execution_status: Optional[List[str]] = _LAST_EXECUTION_STATUS,
    last_modified_by: Optional[List[str]] = _LAST_MODIFIED_BY,
    name: Optional[str] = _LIST_NAME,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** all workflows that exist for the tenant.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak workflows list --page-size 10 --page-number 1 --workflow-status "Draft" --last-execution-status "Success" --last-modified-by "abc@peak.ai" --name "test"
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 25,
        "workflows": [...],
        "workflowsCount": 1
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/get-workflows)
    """
    workflow_client: Workflow = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = workflow_client.list_workflows(
            page_size=page_size,
            page_number=page_number,
            workflow_status=parse_list_of_strings(workflow_status),
            last_execution_status=parse_list_of_strings(last_execution_status),
            last_modified_by=parse_list_of_strings(last_modified_by),
            name=name,
            return_iterator=False,
        )
        writer.write(response)


@app.command(short_help="Describe details of a workflow.")
def describe(
    ctx: typer.Context,
    workflow_id: int = _WORKFLOW_ID,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Describe*** details of a specific workflow.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak workflows describe 9999
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "id": 9999,
        "name": "workflow-name",
        "createdAt": "2020-01-01T18:00:00.000Z",
        "updatedAt": "2020-01-01T18:00:00.000Z",
        "status": "Available",
        "steps": {
            "step-name": {...}
        },
        "lastExecution": {...},
        "triggers": [...],
        "watchers": [...],
        "tags": [...]
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/describe-workflow)
    """
    workflow_client: Workflow = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = workflow_client.describe_workflow(workflow_id=workflow_id)
        writer.write(response)


@app.command(short_help="Delete a workflow.")
def delete(
    ctx: typer.Context,
    workflow_id: int = _WORKFLOW_ID,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Delete*** a workflow.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak workflows delete 9999
    ```

    \b
    🆗 ***Response:***
    ```json
    {}
    ```

    🔗 [**API Documentation**](https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/delete-workflow)
    """
    workflow_client: Workflow = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = workflow_client.delete_workflow(workflow_id=workflow_id)
        writer.write(response)


@app.command(short_help="Start a workflow run.")
def execute(
    ctx: typer.Context,
    workflow_id: int = _WORKFLOW_ID,
    file: Annotated[Optional[str], args.TEMPLATE_PATH] = None,
    params_file: str = args.TEMPLATE_PARAMS_FILE,
    params: List[str] = args.TEMPLATE_PARAMS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Start*** a workflow run. This allows you to pass dynamic parameters to the workflow while running it.

    \b
    🧩 ***Input file schema(yaml):***<br/>
    ```yaml
      body (map):
        params (map | required: false):
            global (map | required: false):
                <keyName> (string | required: false): Key-Value pair. Key is the name of the param which and value is the value of the param.
            stepWise (map | required: false):
                <stepName> (map | required: false): # Parameters to be passed to the step with name `stepName`.
                    <keyName> (string | required: false): Key-Value pair. Key is the name of the param which and value is the value of the param.
        stepsToRun (list(map) | required: false):
            - stepName (string): Name of the step to be executed.
              runDAG (boolean | required: false): Whether to run the dependent steps.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak workflows execute 9999 /path/to/file.yml -v /path/to/params.yml
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "executionId": "d6116a56-6b1d-41b4-a599-fb949f08863f"
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/execute-workflow)
    """
    workflow_client: Workflow = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response: Dict[str, str] = {}

        if file:
            body: Dict[str, Any] = helpers.template_handler(file=file, params_file=params_file, params=params)
            body = helpers.remove_unknown_args(body, workflow_client.execute_workflow)

            response = workflow_client.execute_workflow(workflow_id=workflow_id, **body)
        else:
            response = workflow_client.execute_workflow(workflow_id=workflow_id)

        writer.write(response)


@app.command(short_help="List all available resources.")
def list_resources(
    ctx: typer.Context,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** all available resources that can be used in workflow steps.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak workflows list-resources
    ```

    \b
    🆗 ***Response:***
    ```json
    [
        {
            "cpu": 0.25,
            "gpu": null,
            "gpuMemory": null,
            "id": 21,
            "memory": 0.5
        }
    ]
    ```

    🔗 [**API Documentation**](https://service.peak.ai/workflows/api-docs/index.htm#/Resources/get-resources)
    """
    workflow_client: Workflow = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = workflow_client.list_resources()
        writer.write(response)


@app.command(short_help="List default resources.")
def list_default_resources(
    ctx: typer.Context,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** default resources for the workflows.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak workflows list-default-resources
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "instanceTypeId": 21,
        "storage": "10GB"
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/workflows/api-docs/index.htm#/Resources/get-default-resources)
    """
    workflow_client: Workflow = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = workflow_client.get_default_resource()
        writer.write(response)


@app.command(short_help="List executions for the given workflow.")
def list_executions(
    ctx: typer.Context,
    workflow_id: int = _WORKFLOW_ID,
    date_from: Optional[str] = args.DATE_FROM,
    date_to: Optional[str] = args.DATE_TO,
    status: Optional[List[str]] = _EXECUTION_STATUS,
    count: Optional[int] = _COUNT,
    page_size: Optional[int] = args.PAGE_SIZE,
    page_number: Optional[int] = args.PAGE_NUMBER,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** executions for the given workflow.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak workflows list-executions 9999 --page-size 10 --page-number 1 --status "Running" --status "Failed"
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "executions": [...],
        "pageSize": 25,
        "pageNumber": 1,
        "pageCount": 1,
        "workflowId": 11876
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/workflows/api-docs/index.htm#/Executions/get-workflow-executions)
    """
    workflow_client: Workflow = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = workflow_client.list_executions(
            workflow_id=workflow_id,
            date_from=date_from,
            date_to=date_to,
            status=parse_list_of_strings(status),
            count=count,
            page_size=page_size,
            page_number=page_number,
            return_iterator=False,
        )
        writer.write(response)


@app.command(short_help="Get workflow execution logs.")
def get_execution_logs(
    ctx: typer.Context,
    workflow_id: int = _WORKFLOW_ID_OPTION,
    execution_id: str = _EXECUTION_ID,
    step_name: str = _STEP_NAME,
    next_token: Optional[str] = args.NEXT_TOKEN,
    follow: Optional[bool] = args.FOLLOW,
    save: Optional[bool] = args.SAVE,
    file_name: Optional[str] = args.FILE_NAME,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesOnlyJson] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Get*** execution logs for a specific workflow step.

    \b
    If you want to save the logs to a file, you can use the `--save` flag.
    If you don't provide a file name, the logs will be saved to a file with the name `workflow_execution_logs_<workflow_id>_<execution_id>_<step_name>.log`.

    \b
    If you want to view next set of logs you can pass in the `nextToken` to the command.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak workflows get-execution-logs --workflow-id <workflow_id> --execution-id <execution_id> --step-name <step_name> --next-token <next_token>
    ```

    \b
    To follow the logs, you can use the `--follow` flag.

    \b
    📝 ***Example usage to follow the logs:***<br/>
    ```bash
    peak workflows get-execution-logs --workflow-id <workflow_id> --execution-id <execution_id> --step-name <step_name> --follow
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "stepRunStatus": "Running",
        "logs": [
            {
                "message": "Cloning into service-workflows-api...",
                "timestamp": 1697089188119
            }
        ],
        "nextToken": "f/37846375578651780454915234936364900527730394239380553728/s"
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/workflows/api-docs/index.htm#/Executions/get-workflow-step-logs)
    """
    workflow_client: Workflow = ctx.obj["client"]
    writer = Writer()

    pager = writer.pager()  # Create pager context

    with pager:
        while True:
            response = workflow_client.get_execution_logs(
                workflow_id=workflow_id,
                execution_id=execution_id,
                step_name=step_name,
                next_token=next_token,
                save=save,
                file_name=file_name,
            )

            if save or not response or "stepRunStatus" not in response or "logs" not in response:
                break

            if follow:
                if (response["stepRunStatus"] != "Running") and (
                    response["logs"] is None or len(response["logs"]) == 0
                ):
                    break
                formatted_logs = helpers.format_logs(response["logs"])
                if len(formatted_logs):
                    writer.write(formatted_logs)
            else:
                writer.write(response)
                break

            next_token = response.get("nextToken", None)


@app.command(short_help="Get workflow execution details.")
def get_execution_details(
    ctx: typer.Context,
    workflow_id: int = _WORKFLOW_ID_OPTION,
    execution_id: str = _EXECUTION_ID,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Get*** execution details for a specific workflow run.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak workflows get-execution-details --workflow-id <workflow_id> --execution-id <execution_id>
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "steps": [...],
        "executedAt": "2020-01-01T18:00:00.000Z",
        "finishedAt": "2020-01-01T18:00:00.000Z",
        "runId": "db88c21d-1add-45dd-a72e-8c6b83b68dee",
        "status": "Success"
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/workflows/api-docs/index.htm#/Executions/get-workflow-execution-details)
    """
    workflow_client: Workflow = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = workflow_client.get_execution_details(
            workflow_id=workflow_id,
            execution_id=execution_id,
        )

        writer.write(response)



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#

"""Alerts client module."""

from __future__ import annotations

from typing import Any, Dict, Iterator, List, Literal, Optional, overload

from peak.base_client import BaseClient
from peak.constants import ContentType, HttpMethods
from peak.session import Session


class Alert(BaseClient):
    """Client class for interacting with alerts resource."""

    BASE_ENDPOINT = "notifications/api/v1"

    @overload
    def list_emails(
        self,
        date_from: Optional[str] = None,
        date_to: Optional[str] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        count: Optional[int] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_emails(
        self,
        date_from: Optional[str] = None,
        date_to: Optional[str] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        count: Optional[int] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_emails(
        self,
        date_from: Optional[str] = None,
        date_to: Optional[str] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        count: Optional[int] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Retrieve the history of emails sent.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/notifications/api-docs/index.htm#/emails/get_api_v1_emails>`__

        Args:
            date_from (str | None): The date after which the emails should be included (in ISO format).
                It is 90 days from current date by default.
            date_to (str | None): The date till which the emails should be included (in ISO format).
                It is current date by default.
            page_size (int | None): Number of emails to retrieve per page. It is 25 by default.
            page_number (int | None): Page number to retrieve. Only used when return_iterator is False.
            count (int | None): Number of emails required (Ordered by latest to earliest).
                For example, if 5 is provided, it will return last 5 emails.
                It is -1 by default which means it will return all the available emails within the given dates.
            return_iterator (bool): Whether to return an iterator object or list of emails for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: an iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/emails"

        params: Dict[str, Any] = {
            "dateTo": date_to,
            "dateFrom": date_from,
            "pageSize": page_size,
            "count": -1 if count is None else count,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="emails",
                params=params,
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
        )

    def send_email(
        self,
        body: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Send an email to the specified recipients using the specified template.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/notifications/api-docs/index.htm#/Emails/post_api_v1_emails>`__

        Args:
            body (Dict[str, Any]): A dictionary containing the details of the email to send.

        Returns:
            Dict[str, Any]: A dictionary containing the ID of the email sent.

        SCHEMA:
            .. code-block:: json

                {
                    "recipients": ["string"],
                    "cc": ["string"],
                    "bcc": ["string"],
                    "subject": "string",
                    "templateName": "string",
                    "templateParameters": {
                        "string": "string"
                    }
                }

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given feature does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.BASE_ENDPOINT}/emails"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            body=body,
            content_type=ContentType.APPLICATION_JSON,
        )

    def describe_email(self, email_id: int) -> Dict[str, Any]:
        """Retrieve the details of a specific email.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/notifications/api-docs/index.htm#/Emails/get_api_v1_emails__id_>`__

        Args:
            email_id (int): The ID of the email to retrieve.

        Returns:
            Dict[str, Any]: A dictionary containing the details of the email.

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: Email with the given ID does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/emails/{email_id}"

        return self.session.create_request(endpoint, method, content_type=ContentType.APPLICATION_JSON)  # type: ignore[no-any-return]

    @overload
    def list_templates(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        name: Optional[str] = None,
        scope: Optional[List[str]] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_templates(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        name: Optional[str] = None,
        scope: Optional[List[str]] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_templates(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        name: Optional[str] = None,
        scope: Optional[List[str]] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Retrieve the history of emails sent.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/notifications/api-docs/index.htm#/Templates/get_api_v1_emails_templates>`__

        Args:
            name (str | None): The name of the email template
            scope (List[str] | None): Filter out on the basis of the type of the template - global or custom.
            page_size (int | None): Number of emails to retrieve per page. It is 25 by default.
            page_number (int | None): Page number to retrieve. Only used when return_iterator is False.
            return_iterator (bool): Whether to return an iterator object or list of emails for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: an iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/emails/templates"

        params: Dict[str, Any] = {
            "searchTerm": name,
            "scope": scope,
            "pageSize": page_size,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="templates",
                params=params,
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
        )

    def describe_template(self, template_name: str) -> Dict[str, Any]:
        """Retrieve the details of a specific template.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/notifications/api-docs/index.htm#/Templates/get_api_v1_emails_templates__name_>`__

        Args:
            template_name (str): The name of the template to retrieve.

        Returns:
            Dict[str, Any]: A dictionary containing the details of the template.

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: Email with the given ID does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/emails/templates/{template_name}"

        return self.session.create_request(endpoint, method, content_type=ContentType.APPLICATION_JSON)  # type: ignore[no-any-return]


def get_client(session: Optional[Session] = None) -> Alert:
    """Returns a Alert client, If no session is provided, a default session is used.

    Args:
        session (Optional[Session]): A Session Object. Default is None.

    Returns:
        Alert: The alert client object.
    """
    return Alert(session)


__all__ = ["get_client"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""CLI entrypoint."""
import sys

import peak.config
import typer
from peak.cli import args, helpers
from peak.cli.press import apps, blocks, deployments, specs
from peak.cli.resources import alerts, artifacts, images, services, tenants, users, webapps, workflows
from peak.constants import Sources
from peak.output import Writer

# Workaround to not gray out texts after first paragraph
typer.rich_utils.STYLE_HELPTEXT = ""

typer_app = typer.Typer(rich_markup_mode="markdown", help="Create and Manage Peak Resources")
typer_app.add_typer(images.app, name="images")
typer_app.add_typer(alerts.app, name="alerts")
typer_app.add_typer(artifacts.app, name="artifacts")
typer_app.add_typer(workflows.app, name="workflows")
typer_app.add_typer(webapps.app, name="webapps")
typer_app.add_typer(services.app, name="services")
typer_app.add_typer(tenants.app, name="tenants")
typer_app.add_typer(apps.app, name="apps")
typer_app.add_typer(blocks.app, name="blocks")
typer_app.add_typer(specs.app, name="specs")
typer_app.add_typer(deployments.app, name="deployments")
typer_app.add_typer(users.app, name="users")


@typer_app.callback()
def main(ctx: typer.Context, version: bool = args.VERSION_OPTION) -> None:  # noqa: ARG001
    """Add global version option and add client to context only if any command is invoked."""
    peak.config.SOURCE = Sources.CLI

    if "--help" not in sys.argv and ctx.invoked_subcommand:
        ctx.obj = {
            "client": helpers.get_client(ctx.invoked_subcommand),
            "writer": Writer(),
        }


if __name__ == "__main__":
    typer_app()



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#

"""Peak Alerts commands."""

from typing import Any, Dict, List, Optional

import typer
from peak.cli import helpers
from peak.cli.args import OUTPUT_TYPES, PAGING, TEMPLATE_PARAMS, TEMPLATE_PARAMS_FILE
from peak.constants import OutputTypes, OutputTypesNoTable
from peak.helpers import combine_dictionaries, map_user_options, parse_list_of_strings, variables_to_dict
from peak.output import Writer
from peak.resources.alerts import Alert
from typing_extensions import Annotated

app = typer.Typer(
    help="Alerts management commands.",
    short_help="Manage alerts.",
)

_EMAIL_ID = typer.Argument(..., help="The ID of the email.")
_RECIPIENTS = typer.Option(None, help="The email addresses of the recipients.")
_SUBJECT = typer.Option(None, help="The subject of the email.")
_TEMPLATE_NAME = typer.Option(None, help="The name of the email template.")
_TEMPLATE_PARAMETERS = typer.Option(
    None,
    help="The parameters for the email template. To be passed in stringified JSON format.",
)
_COUNT = typer.Option(None, help="The number of emails to retrieve.")
_DATE_FROM = typer.Option(None, help="The date from which to retrieve emails (in ISO format).")
_DATE_TO = typer.Option(None, help="The date till which to retrieve emails (in ISO format).")
_PAGE_SIZE = typer.Option(None, help="The number of emails per page.")
_PAGE_NUMBER = typer.Option(None, help="The page number to retrieve.")
_CC = typer.Option(None, help="The email addresses of the recipients to be CC'd.")
_BCC = typer.Option(None, help="The email addresses of the recipients to be BCC'd.")
_LIST_TEMPLATE_NAME = typer.Option(None, help="Email Template Name to search for.")
_LIST_TEMPLATE_SCOPE = typer.Option(None, help="List of type of template to filter the templates by.")
_DESCRIBE_TEMPLATE_NAME = typer.Argument(None, help="The name of the email template.")


@app.command("list", short_help="List all emails.")
def list_emails(
    ctx: typer.Context,
    page_size: Optional[int] = _PAGE_SIZE,
    page_number: Optional[int] = _PAGE_NUMBER,
    count: Optional[int] = _COUNT,
    date_from: Optional[str] = _DATE_FROM,
    date_to: Optional[str] = _DATE_TO,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """Retrieve the history of emails sent.

    \b
    📝 ***Example usage:***
    ```bash
    peak alerts emails list --page-size 10 --page-number 1
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "emailCount": 1,
        "emails": [
            {
                "createdAt": "2024-01-01T00:00:00.200Z",
                "createdBy": "platform@peak.ai",
                "id": 1,
                "status": "Delivered",
                "subject": "email_subject",
                "templateName": "template_name",
            }
        ],
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 25
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/notifications/api-docs/index.htm#/Emails/get_api_v1_emails)
    """
    alert_client: Alert = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = alert_client.list_emails(
            page_size=page_size,
            page_number=page_number,
            count=count,
            date_from=date_from,
            date_to=date_to,
            return_iterator=False,
        )
        writer.write(response)


@app.command("send", short_help="Send an email.")
def send_email(
    ctx: typer.Context,
    file: Annotated[
        Optional[str],
        typer.Argument(
            ...,
            help="Path to the file that defines the body for this operation, supports both `yaml` file or a `jinja` template.",
        ),
    ] = None,
    params_file: str = TEMPLATE_PARAMS_FILE,
    params: List[str] = TEMPLATE_PARAMS,
    recipients: Optional[List[str]] = _RECIPIENTS,
    cc: Optional[List[str]] = _CC,
    bcc: Optional[List[str]] = _BCC,
    subject: Optional[str] = _SUBJECT,
    template_name: Optional[str] = _TEMPLATE_NAME,
    template_parameters: Optional[str] = _TEMPLATE_PARAMETERS,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """Send an email to the specified recipients using the specified template.

    \b
    🧩 ***Input file schema(yaml):***<br/>
    ```yaml
      body (map):
        recipients (list(str) | required: true): List of email addresses of the recipients.
        cc (list(str) | required: false): List of email addresses of the recipients to be CC'd.
        bcc (list(str) | required: false): List of email addresses of the recipients to be BCC'd.
        subject (str | required: true): The subject of the email.
        templateName (str | required: true): The name of the email template.
        templateParameters (map | required: false): The parameters for the email template.
    ```

     \b
    📝 ***Example usage:***
    ```bash
    peak alerts emails send '/path/to/email.yaml' --params-file '/path/to/values.yaml'
    ```

    We can also provide the required parameters from the command line or combine the YAML template and command line arguments in which case the command line arguments will take precedence.

    \b
    📝 ***Example usage without yaml:***
    ```bash
    peak alerts emails send --recipients <recipient_email_1> --recipients <recipient_email_2> --subject <email_subject> --template-name <template_name> --template-parameters '{"key": "value"}'
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "id": 1
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/notifications/api-docs/index.htm#/Emails/post_api_v1_emails)
    """
    alert_client: Alert = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    user_options: Dict[str, Any] = variables_to_dict(
        recipients,
        cc,
        bcc,
        subject,
        template_name,
        template_parameters,
    )

    body: Dict[str, Any] = {}
    if file:
        body = helpers.template_handler(file=file, params_file=params_file, params=params)
        body = helpers.remove_unknown_args(body, alert_client.send_email)

    updated_body = combine_dictionaries(body.get("body") or {}, user_options)

    if recipients:
        updated_body["recipients"] = parse_list_of_strings(recipients)

    if cc:
        updated_body["cc"] = parse_list_of_strings(cc)

    if bcc:
        updated_body["bcc"] = parse_list_of_strings(bcc)

    if template_parameters:
        updated_body["templateParameters"] = map_user_options(
            user_options=user_options,
            mapping={},
            dict_type_keys=["templateParameters"],
        )

    with writer.pager():
        response: Dict[str, Any] = alert_client.send_email(body=updated_body)
        writer.write(response)


@app.command(short_help="Describe details of a specific email.")
def describe(
    ctx: typer.Context,
    email_id: int = _EMAIL_ID,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Describe*** details of a specific email.

    \b
    📝 ***Example usage:***
    ```bash
    peak alerts emails describe <email-id>
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "createdAt": "2024-01-01T00:00:00.000Z",
        "createdBy": "platform@peak.ai",
        "deliveredAt": "2024-01-01T00:00:00.000Z",
        "id": 1,
        "recipients": {
            "cc": [],
            "to": ["someone@peak.ai"],
            "bcc": []
        },
        "status": "Delivered",
        "statusDetails": [
            {
                "email": "someone@peak.ai",
                "status": "Delivered",
                "description": "The email was successfully delivered."
            }
        ],
        "subject": "Hello, world!",
        "templateName": "template_name",
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/notifications/api-docs/index.htm#/Emails/get_api_v1_emails__id_)
    """
    alert_client: Alert = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = alert_client.describe_email(email_id=email_id)
        writer.write(response)


@app.command(short_help="List all the templates.")
def list_templates(
    ctx: typer.Context,
    page_size: Optional[int] = _PAGE_SIZE,
    page_number: Optional[int] = _PAGE_NUMBER,
    scope: Optional[List[str]] = _LIST_TEMPLATE_SCOPE,
    name: Optional[str] = _LIST_TEMPLATE_NAME,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """Retrieve the email templates list.

    \b
    📝 ***Example usage:***
    ```bash
    peak alerts emails list-templates --page-size 10 --page-number 1
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "templateCount": 1,
        "templates": [
            {
                "id": 1,
                "createdAt": "2021-09-01T12:00:00Z",
                "createdBy": "platform@peak.ai",
                "name": "test_template",
                "scope": "custom"
            }
        ],
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 25
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/notifications/api-docs/index.htm#/Templates/get_api_v1_emails_templates)
    """
    alert_client: Alert = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = alert_client.list_templates(
            page_size=page_size,
            page_number=page_number,
            scope=scope,
            name=name,
            return_iterator=False,
        )
        writer.write(response)


@app.command(
    short_help="Describe details of a specific template.",
)
def describe_template(
    ctx: typer.Context,
    template_name: str = _DESCRIBE_TEMPLATE_NAME,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Describe*** details of a specific template.

    \b
    📝 ***Example usage:***
    ```bash
    peak alerts emails describe-template <template-name>
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "id": 1,
        "createdAt": "2021-09-01T12:00:00Z",
        "createdBy": "platform@peak.ai",
        "name": "test_template",
        "subject": "Important Account Update Information",
        "body": "<h1>Hello</h1><p>Your account has been updated.</p>",
        "scope": "custom"
    }
    ```

    🔗 [**API Documentation**](https://service.peak.ai/notifications/api-docs/index.htm#/Templates/get_api_v1_emails_templates__name_)
    """
    alert_client: Alert = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = alert_client.describe_template(template_name=template_name)
        writer.write(response)



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Decorator for sending telemetry data for each request."""
from __future__ import annotations

import platform
import threading
import uuid
from contextlib import suppress
from functools import wraps
from typing import Any, Callable, Dict, Optional

import requests

import peak.config
from peak.constants import ContentType, HttpMethods, Stage
from peak.exceptions import BaseHttpException
from peak.helpers import get_base_domain

from ._version import __version__

F = Callable[..., requests.Response]
session_id = str(uuid.uuid4())


def get_status_code(error: Optional[Exception]) -> int | None:
    """It takes an exception object and returns the status code associated with it.

    Args:
        error (Optional[Exception]): The exception object to check

    Returns:
        int | None: The status code related to the error or None if no status code found
    """
    if not error:
        return 200

    if isinstance(error, BaseHttpException):
        return error.STATUS_CODE

    return None


def telemetry(make_request: F) -> F:
    """A decorator that wraps over the make_request function to send telemetry requests as required.

    Args:
        make_request (F): The make_request function to wrap in the decorator

    Returns:
        F: the wrapped function that sends telemetry data for each request
    """

    def get_telemetry_url(session_meta: Optional[Dict[str, Any]] = None) -> str:
        """Returns the telemetry url for the given stage.

        Args:
            session_meta (Optional[Dict[str, Any]]): Session metadata object that contains information like stage

        Returns:
            str: The telemetry URL
        """
        stage = Stage.PROD
        if session_meta:
            stage = session_meta.get("stage", Stage.PROD)
        base_domain = get_base_domain(stage.value, "service")
        return f"{base_domain}/resource-usage/api/v1/telemetry"

    def get_telemetry_data() -> Dict[str, Any]:
        return {
            "sdkVersion": __version__,
            "os": platform.platform(),
            "hostname": platform.uname().node,
            "pythonVersion": platform.python_version(),
            "sessionId": session_id,
            "requestId": str(uuid.uuid4()),
        }

    @wraps(make_request)
    def wrapper(
        self: Any,
        url: str,
        method: HttpMethods,
        content_type: ContentType,
        headers: Optional[Dict[str, str]] = None,
        params: Optional[Dict[str, Any]] = None,
        body: Optional[Dict[str, Any]] = None,
        path: Optional[str] = None,
        request_kwargs: Optional[Dict[str, int | bool | str | float]] = None,
        ignore_files: Optional[list[str]] = None,
        session_meta: Optional[Dict[str, Any]] = None,
    ) -> requests.Response:
        """A decorator that wraps over the make_request function to send telemetry requests as required.

        Args:
            self (Any): the object instance of Handler class on which the make_request call is being made
            url (str): url to send the request to
            method (HttpMethods): The HTTP method to use, e.g. get, post, put, delete
            content_type (ContentType): content type of the request
            headers (Dict[str, str]): headers to send with the request
            params (Dict[str, Any]): params to send to the request
            body (Dict[str, Any]): body to send to the request
            path (str): path to the file or folder that will be compressed and used as artifact, defaults to None
            request_kwargs(Dict[str, int | bool | str | float] | None): extra arguments to be passed when making the request.
            ignore_files(Optional[list[str]]): Ignore files to be used when creating artifact
            session_meta(Dict[str, Any]): Metadata about the session object, like - stage

        Returns:
            requests.Response: response json

        Raises:
            BaseHttpException: The http request failed.
            Exception: Some other error occurred.
        """

        def make_telemetry_request(
            res: Optional[requests.Response] = None,
            error: Optional[Exception] = None,
        ) -> None:
            telemetry_url = get_telemetry_url(session_meta)

            telemetry_body = {
                "response": res.json() if method != HttpMethods.GET and res else None,
                "error": str(error) if error else None,
                "url": url,
                "requestMethod": method.value,
                "statusCode": get_status_code(error),
                "source": peak.config.SOURCE.value,
                **get_telemetry_data(),
            }

            with suppress(Exception):
                make_request(
                    self,
                    telemetry_url,
                    HttpMethods.POST,
                    ContentType.APPLICATION_JSON,
                    headers=headers,
                    body=telemetry_body,
                )

        def trigger_usage_collection(
            res: Optional[Any] = None,
            error: Optional[Exception] = None,
        ) -> None:
            if peak.config.DEBUG_MODE:
                return

            thr = threading.Thread(
                target=make_telemetry_request,
                kwargs={
                    "res": res,
                    "error": error,
                },
            )

            thr.start()

        try:
            custom_headers = {f"x-peak-{key}": value for (key, value) in get_telemetry_data().items()}
            custom_headers = {
                **custom_headers,
                **(headers or {}),
            }

            res = make_request(
                self,
                url,
                method,
                content_type=content_type,
                headers=custom_headers,
                params=params or {},
                body=body or {},
                path=path,
                ignore_files=ignore_files,
                request_kwargs=request_kwargs,
            )

            trigger_usage_collection(res=res)
        except Exception as e:
            trigger_usage_collection(error=e)
            raise
        else:
            return res

    return wrapper



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Exceptions for the Peak API."""

from __future__ import annotations

from collections import defaultdict
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Optional, Tuple, Type


class PeakBaseException(Exception):
    """Base exception class for the Peak SDK."""


class HttpExceptionsRegistryMeta(type):
    """Registry metaclass."""

    REGISTRY: ClassVar[Dict[int, Any]] = defaultdict(lambda: Exception)

    def __new__(
        cls: "Type[HttpExceptionsRegistryMeta]",
        name: str,
        bases: Tuple[Any, ...],
        attrs: Dict[str, Any],
    ) -> HttpExceptionsRegistryMeta:
        """This method runs whenever a new class (that uses this class as its metaclass) is defined.

        This method automatically adds the exception classes to its Registry.
        It uses the `STATUS_CODE` attribute of the class as key and the class itself as value in registry.
        Ref: https://charlesreid1.github.io/python-patterns-the-registry.html

        Args:
            name (str): Name of the child class
            bases (tuple): Tuple of the child class's inheritance tree
            attrs (dict): Name and value pairs of all the attributes defined in the child class

        Returns:
            HttpExceptionsRegistryMeta: the child class itself, forward annotated for type checking
        """
        new_cls: "HttpExceptionsRegistryMeta" = type.__new__(cls, name, bases, attrs)
        status_code: Optional[int] = attrs.get("STATUS_CODE")
        if status_code:
            cls.REGISTRY[status_code] = new_cls
        return new_cls


class BaseHttpException(PeakBaseException, metaclass=HttpExceptionsRegistryMeta):
    """Base registry class for registering all exceptions."""

    STATUS_CODE: ClassVar[int]


class BadRequestException(BaseHttpException):
    """The provided inputs are invalid."""

    STATUS_CODE = 400


class UnauthorizedException(BaseHttpException):
    """The authentication credentials are invalid or expired."""

    STATUS_CODE = 401


class ForbiddenException(BaseHttpException):
    """User does not have permissions to perform the operation."""

    STATUS_CODE = 403


class NotFoundException(BaseHttpException):
    """Resource does not exist."""

    STATUS_CODE = 404


class ConflictException(BaseHttpException):
    """There is a conflict with the current state of the target resource."""

    STATUS_CODE = 409


class PayloadTooLargeException(BaseHttpException):
    """The provided file size is larger than the maximum limits."""

    STATUS_CODE = 413

    def __init__(self, message: str = "") -> None:
        """Throw exception with custom message.

        Args:
            message (str): Additional message to add to exception.
        """
        error_message: str = (
            "Please use '--dry-run' with your command to preview the file tree and optimize the file contents."
        )
        super().__init__(f"{message}. {error_message}")


class UnprocessableEntityException(BaseHttpException):
    """The server understands the request, but it was unable to process the contained instructions."""

    STATUS_CODE = 422


class InternalServerErrorException(BaseHttpException):
    """The server encountered an unexpected condition that prevented it from fulfilling the request."""

    STATUS_CODE = 500


class InvalidPathException(PeakBaseException):
    """The provided path is invalid and cannot be processed."""

    def __init__(self, path: str | Path, message: str = "") -> None:
        """Throw exception with custom message.

        Args:
            path (str | Path): Path which is invalid.
            message (str): Any extra message to add to exception.
        """
        super().__init__(f"Invalid path: {path!r}. {message}")


class MissingEnvironmentVariableException(PeakBaseException):
    """Required environment variable not found."""

    def __init__(self, env_var: str, *, message: str = "") -> None:
        """Throw exception with custom message.

        Args:
            env_var (str): Name of env variable which is not present.
            message (str): Any extra message to add to exception.
        """
        error_message: str = f"{env_var} environment variable is not set or is empty."
        super().__init__(f"{error_message} {message}")


class FileLimitExceededException(PeakBaseException):
    """Limits on the file are exceeded."""

    def __init__(self, max_size: float, *, message: str = "", units: str = "MB") -> None:
        """Throw exception with custom message.

        Args:
            max_size (float): Maximum size of the file.
            message (str): Additional message to add to exception.
            units (str): Units of the maximum size.
        """
        error_message: str = (
            f"Compressed directory size is over {max_size}{units}. Please use '--dry-run' with your command to preview the file tree and optimize the file contents."
        )
        super().__init__(f"{error_message} {message}")


class InvalidTemplateException(PeakBaseException):
    """The given template is invalid and could not be compiled."""

    def __init__(self, message: str) -> None:
        """Throw exception with custom message.

        Args:
            message (str): Message of the exception.
        """
        super().__init__(message)


class BadParameterException(PeakBaseException):
    """Raises exception for invalid parameters."""

    def __init__(self, param: str, *, message: str = "") -> None:
        """Throw exception with custom message.

        Args:
            param (str): Raw parameter input from the CLI.
            message (str): Additional message to add to exception.
        """
        error_message: str = f"Unable to parse: {param}"
        super().__init__(f"{error_message} {message}")


class InvalidParameterException(PeakBaseException):
    """Raises exception for invalid parameters."""

    def __init__(self, *, message: str = "") -> None:
        """Throw exception with custom message.

        Args:
            message (str): Additional message to add to exception.
        """
        super().__init__(message)


__all__: List[str] = [
    "HttpExceptionsRegistryMeta",
    "BaseHttpException",
    "BadRequestException",
    "UnauthorizedException",
    "ForbiddenException",
    "NotFoundException",
    "ConflictException",
    "PayloadTooLargeException",
    "UnprocessableEntityException",
    "InternalServerErrorException",
    "InvalidPathException",
    "MissingEnvironmentVariableException",
    "FileLimitExceededException",
    "BadParameterException",
    "InvalidParameterException",
]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""This file is part of the peak-sdk framework.

Copyright © 2024 Peak AI

Unauthorized copying of this file, via any medium is strictly prohibited.
"""
from __future__ import annotations

from typing import List

from peak import press, resources, tools
from peak.session import Session

from ._version import __version__

__all__: List[str] = ["Session", "press", "resources", "tools", "__version__"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Logging module, a wrapper around `structlog <https://www.structlog.org/en/stable/>`_ library."""

from __future__ import annotations

import functools
import inspect
import logging
import os
import sys
from types import MappingProxyType
from typing import Any, Callable, Final, Hashable, List, MutableMapping, Optional, Tuple, Union

import orjson
import structlog

from .log_handler import LogHandler
from .log_level import LOG_LEVEL_NAMES_TO_LOG_LEVEL, LogLevel, LogLevelNames
from .utils import mask_nested_pii_data

__title__ = "logging"
__author__ = "PEAK AI"
__license__ = "Apache License, Version 2.0"
__copyright__ = "2024, Peak AI"
__status__ = "production"
__date__ = "14 March 2024"

__all__: List[str] = [
    "DEFAULT_SHARED_PROCESSORS",
    "LOG_LEVEL_NAMES_TO_LOG_LEVEL",
    "LogHandler",
    "LogLevelNames",
    "LogLevel",
    "PeakLogger",
    "default_processors_factory",
    "get_logger",
    "pii_masking_processor",
    "peak_contexts_processor",
]


# ---------------------------------------------------------------------------
# Utility private functions
# ---------------------------------------------------------------------------


def pii_masking_processor(
    _: str,
    __: str,
    event_dict: MutableMapping[str, Any],
) -> MutableMapping[str, Any]:
    """Masks sensitive PII data present in event_dict."""
    return mask_nested_pii_data(event_dict)


def peak_contexts_processor(
    _: str,
    __: str,
    event_dict: MutableMapping[str, Any],
) -> MutableMapping[str, Any]:
    """Add the standard attribute to the event_dict."""
    attributes_to_add: dict[str, Any] = {
        "source": "peak-sdk",
        "runtime": os.getenv("PEAK_RUNTIME"),
        "press_deployment_id": os.getenv("PRESS_DEPLOYMENT_ID"),
        "run_id": os.getenv("PEAK_RUN_ID"),
        "exec_id": os.getenv("PEAK_EXEC_ID"),
        "stage": os.getenv("STAGE"),
        "tenant_name": os.getenv("TENANT_NAME", os.getenv("TENANT")),
        "tenant_id": os.getenv("TENANT_ID"),
        "api_name": os.getenv("PEAK_API_NAME"),
        "api_id": os.getenv("PEAK_API_ID"),
        "step_name": os.getenv("PEAK_STEP_NAME"),
        "step_id": os.getenv("PEAK_STEP_ID"),
        "webapp_name": os.getenv("PEAK_WEBAPP_NAME"),
        "webapp_id": os.getenv("PEAK_WEBAPP_ID"),
        "workflow_name": os.getenv("PEAK_WORKFLOW_NAME"),
        "workflow_id": os.getenv("PEAK_WORKFLOW_ID"),
        "workspace_name": os.getenv("PEAK_WORKSPACE_NAME"),
        "workspace_id": os.getenv("PEAK_WORKSPACE_ID"),
        "image_name": os.getenv("PEAK_IMAGE_NAME"),
        "image_id": os.getenv("PEAK_IMAGE_ID"),
    }

    for attr, value in attributes_to_add.items():
        if value:
            event_dict[attr] = value

    return event_dict


# ---------------------------------------------------------------------------
# Utility functions at module level for main logger factory
# ---------------------------------------------------------------------------


DEFAULT_SHARED_PROCESSORS: Tuple[structlog.types.Processor | Any, ...] = (
    structlog.contextvars.merge_contextvars,
    peak_contexts_processor,
    structlog.stdlib.filter_by_level,
    structlog.stdlib.add_logger_name,
    structlog.stdlib.add_log_level,
    structlog.stdlib.PositionalArgumentsFormatter(),
    pii_masking_processor,
    structlog.processors.TimeStamper(fmt="iso", utc=True),
    structlog.processors.StackInfoRenderer(),
    structlog.processors.format_exc_info,
    structlog.processors.UnicodeDecoder(),
    structlog.processors.EventRenamer("message"),
)

_ORJSON_OPTS: Final[int] = (
    orjson.OPT_SERIALIZE_NUMPY
    | orjson.OPT_SERIALIZE_DATACLASS
    | orjson.OPT_SERIALIZE_UUID
    | orjson.OPT_NON_STR_KEYS
    | orjson.OPT_SORT_KEYS
)


def _orjson_serializer(
    obj: Any,
    sort_keys: Optional[bool] = None,
    default: Callable[[Any], Any] = str,
) -> str:
    """Custom serializer using orjson.dumps for structlog."""
    apply_opts: int = (_ORJSON_OPTS | orjson.OPT_SORT_KEYS) if sort_keys else _ORJSON_OPTS

    return orjson.dumps(obj, option=apply_opts, default=default).decode("utf-8")


@functools.lru_cache(maxsize=2, typed=True)  # Only 2 different combinations of disable_masking are possible
def default_processors_factory(
    disable_masking: Optional[bool],
) -> list[structlog.types.Processor | Any]:
    """Return the default processors for PeakLogger.

    Args:
        disable_masking (Optional[bool], optional): Whether to disable masking of sensitive data. Defaults to False.

    Returns:
        list[structlog.types.Processor | Any]: List of processors to be used by the logger.
    """
    _processors = list(DEFAULT_SHARED_PROCESSORS)

    if disable_masking:
        _processors.remove(pii_masking_processor)

    # add renderer based on the environment
    if sys.stdout.isatty():
        # Pretty printing when we run in a terminal session.
        _processors.remove(structlog.processors.format_exc_info)
        _processors.append(
            structlog.dev.ConsoleRenderer(
                colors=True,
                event_key="message",
                timestamp_key="timestamp",
                exception_formatter=structlog.dev.RichTracebackFormatter(color_system="truecolor"),
            ),
        )
    else:
        # Print JSON when we run in production
        _processors.append(structlog.processors.JSONRenderer(serializer=_orjson_serializer, sort_keys=True))

    return _processors


@functools.lru_cache(maxsize=128, typed=True)
def _handle_and_patch_processor_factory_kwargs(
    func: Callable[..., List[structlog.types.Processor | Any]],
    **kwargs: Hashable,
) -> List[structlog.types.Processor | Any]:
    """Handle keyword arguments for custom_processors_factory using inspect.signature, additionally patch the processors list to include EventRenamer in the right position if not already present.

    Unknown keyword arguments are ignored.

    Args:
        func (Callable[..., List[structlog.types.Processor | Any]]): Custom processor factory function.
        **kwargs: Additional keyword arguments to be passed to the custom_processors_factory, if provided.
            kwargs received by the factory function must be hashable else TypeError will be raised by this wrapper.

    Returns:
        List[structlog.types.Processor | Any]: List of processors to be used by the logger.

    Raises:
        ValueError: If multiple renderers are found in the processor factory's returned processors list.
    """
    func_params: MappingProxyType[str, inspect.Parameter] = inspect.signature(func).parameters
    _processors = func(**{k: v for k, v in kwargs.items() if k in func_params})

    if "structlog.processors.EventRenamer" not in str(_processors):

        # find index of KeyValueRenderer/JSONRenderer/ConsoleRenderer and push EventRenamer to before either of them
        indices_for_insertion: list[int] = [
            _processors.index(processor)
            for processor in _processors
            if getattr(processor, "__name__", processor.__class__.__name__)
            in ("KeyValueRenderer", "JSONRenderer", "ConsoleRenderer")
        ]

        if len(indices_for_insertion) > 1:
            multiple_renderer_error_msg: str = f"""
            Multiple renderers found in the processors list returned by the `custom_processors_factory` function: {func.__name__}.
            Please ensure only one of KeyValueRenderer, JSONRenderer, or ConsoleRenderer is present in the processors list.
            """
            raise ValueError(multiple_renderer_error_msg)

        _processors.insert(
            min([*indices_for_insertion, len(_processors)]),
            structlog.processors.EventRenamer("message"),
        )

    return _processors


# ---------------------------------------------------------------------------
# Logger factory function
# ---------------------------------------------------------------------------


def get_logger(
    name: Optional[str] = None,
    level: Optional[LogLevel] = LogLevel.INFO,
    custom_processors_factory: Optional[Callable[..., List[structlog.types.Processor | Any]]] = None,
    disable_masking: Optional[bool] = False,  # noqa: FBT002
    handlers: Optional[List[LogHandler]] = None,
    file_name: Optional[str] = None,
    **kwargs: Any,
) -> PeakLogger:
    """Return a logger with the specified settings.

    When using the default implementation, pretty-printing is automatically enabled when logger is run in a terminal session (sys.stdout.isatty() == True)
    and JSON printing is enabled when logger is run in production via the `structlog.processors.JSONRenderer` processor.

    Args:
        name (Optional[str], optional): Name of the logger. Defaults to None.
        level (LogLevel): Log level. Defaults to LogLevel.INFO.
        custom_processors_factory (Optional[Callable[..., List[structlog.types.Processor | Any]]], optional): A factory function that returns a list of custom processors.
            Defaults to None. This disables the default processors provided with the default implementation.
        disable_masking (Optional[bool], optional): Whether to disable masking of sensitive data. Defaults to False.
            Only applicable when using the default processors, as custom processors can be used to handle masking on their own.
        handlers (Optional[List[Handlers]], optional): List of log handlers (CONSOLE, FILE). Defaults to CONSOLE.
        file_name (Optional[str], optional): Filename for FILE handler. Required if FILE handler is used. Defaults to None.
        **kwargs: Additional keyword arguments to be passed to the custom_processors_factory, if provided.
            `disable_masking` is automatically passed to the custom_processors_factory and should not be provided here.
            if `custom_processors_factory` does not accept any keyword arguments, they will all be ignored.
            Additionally, all kwargs receivable by the factory function must be hashable else TypeError will be raised by the `_handle_and_patch_processor_factory_kwargs` wrapper.

    Returns:
        PeakLogger: A logger instance configured with the specified settings.

    Raises:
        ValueError: If the `file_name` is not provided for FILE handler or if `multiple renderers` are found in the `processor`(s) list returned by the `custom_processors_factory`.
    """
    _log_level: int = (
        level.value
        if level is not None
        else logging.DEBUG if os.getenv("DEBUG", "false").lower() == "true" else LogLevel.INFO.value
    )
    _processors: list[structlog.types.Processor | Any] = (
        _handle_and_patch_processor_factory_kwargs(custom_processors_factory, disable_masking=disable_masking, **kwargs)
        if custom_processors_factory is not None
        else default_processors_factory(
            disable_masking=disable_masking,
        )
    )
    handlers_list: list[Any] = []
    if not handlers or LogHandler.CONSOLE in handlers:
        handlers_list.append(logging.StreamHandler())  # Console handler
    if handlers and LogHandler.FILE in handlers:
        if file_name:
            handlers_list.append(logging.FileHandler(file_name))  # File handler
        else:
            msg = "filename must be provided for FILE handler."
            raise ValueError(msg)

    # Set the log level and add the handlers to the root logger
    # This is required to ensure that the log level and handlers are applied to the root logger
    logging.basicConfig(level=_log_level, handlers=handlers_list, format="", force=True)

    # configure structlog with the specified settings
    structlog.configure(
        processors=[*_processors],
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

    return PeakLogger(structlog.get_logger(name))


# ---------------------------------------------------------------------------
# Wrapper Logger class
# Basically delegate everything to `structlog`.
# ---------------------------------------------------------------------------


class PeakLogger:
    """Wrapper class for logging with various log levels."""

    # use __slots__ to avoid dynamic attribute creation
    __slots__: list[str] = ["_logger"]

    def __init__(self, logger: Any) -> None:
        """Initialize with a logger object.

        Args:
            logger (Any): Logger object to wrap.
        """
        self._logger: structlog.stdlib.BoundLogger = logger

    def __getattribute__(self, __name: str) -> Any:
        """Return the attribute from the wrapped logger object."""
        if __name in [*PeakLogger.__slots__, *PeakLogger.__dict__.keys()]:
            return object.__getattribute__(self, __name)
        return getattr(self._logger, __name)

    def bind(self, context: Union[dict[str, Any], None] = None, **kwargs: Any) -> None:
        """Bind contextual information to the logger, enriching log messages.

        This method allows attaching context data to the logger, such as additional information
        or system details, to provide more context in log messages.

        Args:
            context (Union[dict[str, Any], None]): A dictionary or None for contextual information.
            **kwargs: Additional key-value pairs to enhance context.
        """
        if context is None:
            context = {}

        if kwargs:
            # file deepcode ignore AttributeLoadOnNone: false positive
            context.update(kwargs)

        self._logger = self._logger.bind(**context)

    def unbind(self, keys: list[str]) -> None:
        """Unbind specified keys from the logger's context.

        Args:
            keys (list[str]): List of keys to unbind.
        """
        context: dict[str, Any] | dict[Any, Any] = structlog.get_context(self._logger)

        for key in keys:
            if key in context:
                del context[key]

        # Rebind the modified context to the logger
        self._logger = self._logger.bind(**context)

    def clone_with_context(self, context: Union[dict[str, Any], None] = None, **kwargs: Any) -> PeakLogger:
        """Return a frozen copy of this logger with the specified context added."""
        new_logger = PeakLogger(self._logger.new())
        new_logger.bind(context, **kwargs)
        return new_logger

    def set_log_level(self, level: LogLevel) -> None:
        """Set the log level of the root logger.

        Args:
            level (LogLevel): Log level to set.
        """
        if self._is_valid_log_level(level):
            logging.getLogger().setLevel(level.value)

    def _is_valid_log_level(self, level: LogLevel) -> bool:
        """Check if a given log level is valid."""
        return level in LogLevel



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#

"""Collection of basic helper functions."""

from __future__ import annotations

import inspect
import json
import re
from datetime import datetime, timezone
from types import FrameType
from typing import Any, Dict, List, Optional

import requests

from peak import config
from peak.constants import Sources
from peak.output import Writer


def parse_body_for_multipart_request(body: Dict[str, Any]) -> Dict[str, str]:
    """Parses an object to make it suitable for passing in a multipart request.

    Args:
        body (Dict[str, Any]): the object to be parsed

    Returns:
        Dict[str, str]: the parsed object
    """
    return {key: (value if isinstance(value, str) else json.dumps(value)) for (key, value) in body.items()}


def remove_keys(body: Dict[str, Any], keys: List[str]) -> Dict[str, Any]:
    """Removes given keys from a dictionary.

    Args:
        body (Dict[str, Any]): the object to be parsed
        keys (List[str]): the keys to remove

    Returns:
        Dict[str, str]: the final object with required keys removed
    """
    return {key: value for (key, value) in body.items() if key not in keys}


def get_base_domain(stage: str, subdomain: Optional[str] = "service") -> str:
    """Gets the base domain for a stage with the given subdomain.

    Args:
        stage (str): the stage
        subdomain (Optional[str]): the subdomain

    Returns:
        str: the final base domain
    """
    if stage == "prod":
        stage = ""
    elif stage == "latest":
        stage = "dev"

    domain: str = f"https://{subdomain}.{stage}.peak.ai"
    return domain.replace("..", ".")  # for prod domain


def parse_list_of_strings(param: List[str] | None) -> List[str] | None:
    """Split comma separated strings in the list and flatten that list.

    Args:
       param (List[str] | None): List of strings

    Returns:
        List[Any] | None: The final flattened list
    """
    if param is None or len(param) == 0:
        return param

    result: List[str] = []
    for e in param:
        result = result + e.split(",")

    return result


def snake_case_to_lower_camel_case(snake_case_string: str) -> str:
    """Converts underscore string to lower camel case.

    Args:
        snake_case_string (str): string in underscore

    Returns:
        str: lower camel case string
    """
    parts = snake_case_string.split("_")
    return parts[0] + "".join(part.capitalize() for part in parts[1:])


def variables_to_dict(*args: Any, frame: FrameType | None = None) -> Dict[str, str]:
    """Converts arbitrary variables to a dictionary.

    Args:
        args (str|int): tuple of string|int variables
        frame (FrameType|None): Current Frame of caller

    Returns:
        Dict[str, str]: Dictionary containing key value pair of variables
    """
    if frame is None:
        frame = inspect.currentframe()
        if frame:
            # set frame to previous
            frame = frame.f_back

    var_dict = {}
    if frame and frame.f_locals:
        for var_name, var_value in frame.f_locals.items():
            if var_value in args and var_value:
                var_dict[snake_case_to_lower_camel_case(var_name)] = var_value

    del frame  # Explicitly release the frame to avoid reference cycles
    return var_dict


def combine_dictionaries(
    dict1: Dict[str, Any],
    dict2: Dict[str, Any],
    nested_keys_to_skip: Optional[List[str]] = [],  # noqa: B006
) -> Dict[str, Any]:
    """Combines two dictionaries. Values for second dictionary have higher precedence.

    Args:
        dict1 (Dict[str, Any]): dictionary 1
        dict2 (Dict[str, Any]): dictionary 2
        nested_keys_to_skip (List[str] | None): Keys for which nested combining is not required.

    Returns:
        Dict[str, Any]: Combined dictionary
    """
    if not dict1:
        return dict2

    combined_dict = dict(dict1)
    for key in dict2:
        if key in combined_dict and isinstance(combined_dict[key], dict) and key not in (nested_keys_to_skip or []):
            combined_dict[key] = combine_dictionaries(combined_dict[key], dict2[key])
        else:
            combined_dict[key] = dict2[key]
    return combined_dict


def map_user_options(
    user_options: Dict[str, Any],
    mapping: Dict[str, str],
    dict_type_keys: List[str] = [],  # noqa: B006
) -> Dict[str, Any]:
    """Maps user provided inputs to a specific format.

    Args:
        user_options (Dict[str, Any]): Dictionary containing user inputs
        mapping (Dict[str, Any]): Mapping to be used for conversion
        dict_type_keys (List[str]): List of keys which have json type values

    Returns:
        Dict[str, str]: Mapped dictionary
    """
    result: Dict[str, Any] = {}
    for key in user_options:
        if key in mapping:
            nested_dict = result.get(mapping[key], {})
            nested_dict[key] = json.loads(user_options[key]) if key in dict_type_keys else user_options[key]
            result[mapping[key]] = nested_dict
        else:
            result[key] = json.loads(user_options[key]) if key in dict_type_keys else user_options[key]
    return result


def remove_none_values(data: Any) -> Dict[str, Any]:
    """Recursively remove all keys with `None` values from a dictionary.

    Args:
        data (Any): Dictionary to be cleaned.

    Returns:
        Dict[str, Any]: Dictionary with all keys with `None` values removed.
    """
    if isinstance(data, dict):
        return {k: remove_none_values(v) for k, v in data.items() if v is not None}

    return data  # type: ignore[no-any-return]


def convert_to_snake_case(s: str) -> str:
    """Converts a given string from camelCase or TitleCase to snake case.

    Args:
        s (str): The string to be converted.

    Returns:
        str: The converted string in snake case.
    """
    return re.sub(r"(?<!^)(?=[A-Z])", "_", s).lower()


def convert_keys_to_snake_case(data: Dict[str, Any], *, convert_nested: bool = False) -> Dict[str, Any]:
    """Converts keys of a dictionary to snake case.

    Args:
        data (Dict[str, Any]): Dictionary to be converted.
        convert_nested (bool): Whether to convert nested keys as well. Default is False.

    Returns:
        Dict[str, Any]: Dictionary with keys converted to snake case.
    """

    def convert_dict(d: Dict[str, Any]) -> Dict[str, Any]:
        new_dict = {}
        for k, v in d.items():
            new_key = convert_to_snake_case(k)
            if convert_nested and isinstance(v, dict):
                new_dict[new_key] = convert_dict(v)
            else:
                new_dict[new_key] = v
        return new_dict

    return convert_dict(data)


def format_date(timestamp: str, time_format: str = "%Y/%m/%d %H:%M:%S") -> str:
    """Format a timestamp to a given format.

    Args:
        timestamp (str): Timestamp to be formatted in milliseconds.
        time_format (str): Format to be used for formatting the given timestamp.

    Returns:
        str: Formatted timestamp.
    """
    return datetime.fromtimestamp(int(timestamp) / 1000, tz=timezone.utc).strftime(time_format)


def download_logs_helper(
    response: Dict[str, Any],
    file_name: str,
) -> None:
    """Save logs to a file or return the download url.

    Args:
        response (Dict[str, Any]): Response object from the API.
        file_name (str): Name of the file where the contents should be saved.

    Returns: None
    """
    download_response = requests.get(response["downloadUrl"], timeout=20, stream=True)
    download_response.raise_for_status()

    with open(file_name, "wb") as file:  # noqa: PTH123
        file.write(download_response.content)
        if config.SOURCE == Sources.CLI:
            writer = Writer()
            writer.write(f"\nLog contents have been saved to: {file_name}")


def search_action(current_path: str, current_dict: Dict[str, Any], action: str) -> Dict[str, bool]:
    """Search for a specified action within a nested dictionary structure and check if deeper levels exist.

    Args:
        current_path (str): The dot-separated path representing the feature hierarchy.
        current_dict (Dict[str, Any]): The nested dictionary representing the permissions structure.
        action (str): The action to search for (e.g., "read" or "write").

    Returns:
        bool: A dictionary with keys 'has_permission' indicating if the action is found and
              'deeper_levels' indicating if there are deeper levels that need specification.
    """
    keys = current_path.split(".")
    for key in keys:
        if key in current_dict:
            current_dict = current_dict[key]
        else:
            return {"has_permission": False, "deeper_levels": False}

    if "actions" in current_dict:
        actions = current_dict["actions"]
        has_permission = "*" in actions or action in actions or ("write" in actions and action == "read")
        return {"has_permission": has_permission, "deeper_levels": False}

    return {"has_permission": False, "deeper_levels": True}
